/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: music_midi.c,v 1.7 2000/08/12 10:32:56 tf Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <signal.h>
#include <sys/wait.h>
#include <string.h>

#ifndef True
#include <X11/Xlib.h>
#endif

#include "hana.h"

#define TABLE_POS 0x00000960
#define OFFSET (0x00000bf0 - 0x17f0)
#define MUSIC_NUM 40

static int midi_prepare(int, const char *);
static int midi_exec_player(const char *);
static int music_midi_play(int);
static int music_midi_stop(int);
static void midi_wait_child(void);

#if 0
int main(int argc, char *argv[])
{
  int fd;
  char buf[BUFSIZ];
  int i;

  fd =  music_midi_process();
  while(1) {
    i = read(0, buf, BUFSIZ - 1);
    write(fd, buf, i);
  }
}
#endif

int music_midi_process()
{
  int pid;
  int fd[2];
  FILE *fp;
  int track;
  char buf[BUFSIZ];
  int arg;
  Bool loop;

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#define PF_LOCAL PF_UNIX
#endif
  if (socketpair(AF_LOCAL, SOCK_STREAM, PF_LOCAL, fd) < 0) {
    perror("socketpair AF_LOCAL/PF_LOCAL");
    if (socketpair(AF_UNIX, SOCK_STREAM, AF_UNSPEC, fd) < 0) {
      perror("socketpair AF_LOCAL/AF_UNSPEC");
      return -1;
    }
  }

  if ((pid = fork()) < 0) {
    perror("fork");
    close(fd[0]);
    close(fd[1]);
    return -1;
  } else if (pid) {
    close(fd[1]);
    return  fd[0];
  }
  
  fprintf(stderr, "started MIDI control process.\n");
  close(fd[0]);
  
  if ((fp = fdopen(fd[1], "rw")) == NULL) {
    perror("fdopen");
    exit(1);
  }

  pid = -1;
    
  while (fgets(buf, sizeof(buf), fp)) {
    char *p;
    if ((p = strchr(buf, '\n'))) {
      *p = '\0';
    }
    switch(buf[0]) {
    case 't': /* play track */
      sscanf(buf + 1, "%d", &track);
      printf("play track %d\n", track);
      pid = music_midi_play(track);
      loop = False;
      break;
    case 'l': /* play track with loop */
      sscanf(buf + 1, "%d", &track);
      printf("play track %d with loop\n", track);
      pid = music_midi_play(track);
      loop = True;
      break;
    case 'p': /* pause */
      // music_cd_pause(c);
      // pause_loop = loop;
      loop = False;
      break;
    case 'r': /* resume*/
      // music_cd_resume(c);
      // loop = pause_loop;
      break;
    case 's': /* stop */
      music_midi_stop(pid);
      loop = False;
      break;
    case 'f': /* fadeout */
      sscanf(buf + 1, "%d", &arg);
      printf("fadeout in %d msec.\n", arg);
      // music_cd_stop_fadeout(c, arg);
      break;
    case 'v': /* set volume */
      sscanf(buf + 1, "%d", &arg);
      printf("set volume to %d.\n", arg);
      // music_cd_set_volume(c, arg);
      break;
    case 'x': /* exit */
      goto exit;
    }
  }
exit:
  
  exit(0);
}

static int music_midi_play(int id)
{
  char fname[BUFSIZ];
  int pid;
  
  if (id > MUSIC_NUM) {
    fprintf(stderr, "invalid music file is specified.\n");
    return -1;
  }
  snprintf(fname, BUFSIZ, "%s/0x%x.mid", path.midiwork, id);

  midi_prepare(id, fname);
  pid = midi_exec_player(fname);
  signal(SIGCHLD, (void *)midi_wait_child);

  return pid;
}

static int music_midi_stop(int pid)
{
  if (pid > 0) {
    kill(pid, SIGTERM);
  }
  return 0;
}

static int midi_exec_player(const char *fname)
{
  int pid;

  pid = fork();
  if (pid < 0) {
    perror("fork");
    return -1;
  } else if (pid) { /* parent */
    signal(SIGCHLD, (void *)midi_wait_child);
    return pid;
  }
  execlp(path.midiplayer, "hanamus", fname, NULL);

  return 0;
}

static int midi_prepare(int id, const char *fname)
{
  FILE *out;
  int fd;
  long p;
  unsigned char *addr;
  struct stat sb;
  int position, size;
  char midipath[PATHLEN];

  if ((fd = open(fname, 0, O_RDONLY)) >= 0) {
    fprintf(stderr, "%s already exists.\n", fname);
    return 0;
  }
  
  snprintf(midipath, PATHLEN, "%s/hanamus.dll", path.orig);
  if ((fd = open(midipath, 0, O_RDONLY)) < 0) {
    perror("midi_prepare: open");
    return -1;
  }
  if (fstat(fd, &sb) < 0 ||
      (addr = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0))
      == MAP_FAILED) {
    perror("mmap");
    close(fd);
    exit(1);
  }
  close(fd);

  /* read table */
  p = TABLE_POS + 16 * (id - 2);
  position = (addr[p + 3] << 24) | (addr[p + 2] << 16) | (addr[p + 1] << 8) | addr[p];
  p += 4;
  size = (addr[p + 3] << 24) | (addr[p + 2] << 16) | (addr[p + 1] << 8) | addr[p];
  
  /* read and output MIDI data */
  // fprintf(stderr, "%x:%x:%d\n", id, position, size);
  fprintf(stderr, "output %s...", fname);
  out = fopen(fname, "w");
  fwrite(&addr[position + OFFSET], 1, size, out);
  fprintf(stderr, "done.\n");
  fclose(out);
  
  return 0;
}

static void midi_wait_child()
{
  int pid, status;

  pid = wait(&status);
}
