/*
 * png.c (for Hanashizume no Matsuri for X)
 * Copyright (c) 2000 Kimihiro Nonaka <aw9k-nnk@asahi-net.or.jp>
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by Kimihiro Nonaka and TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: png.c,v 1.3 2000/12/12 18:14:13 tf Exp $
 */

#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <png.h>

#include "hana.h"

static void write_png(Image *, const char *);

#ifndef png_jmpbuf
#define png_jmpbuf(pngp) (pngp->jmpbuf)
#endif

void
output_omake()
{
  RESINFO *r;
  unsigned char *p;
  int size = 1770477;
  char name[] = "CGHANA_PNG";
  char filename[MAX_PATH];
  FILE *out;

  strncpy(filename, path.imagedir, MAX_PATH);
  strncat(filename, "/CGHANA_PNG.png", MAX_PATH);

  r = search_rlist(cg.rlist, name);
  if (r == NULL) {
    fprintf(stderr, "Can't find the image: %s.\n", name);
    exit(1);
  }
  
  p = (unsigned char *)calloc(size, sizeof(unsigned char));
  if (p == NULL) {
    fprintf(stderr, "output_omake: Can't allocate memory.\n");
    exit(1);
  }
  fseek(cg.fd, r->offset, SEEK_SET);
  fread(p, 1, size, cg.fd);
  out = fopen(filename, "w");
  if (out != NULL) {
    fwrite(p, 1, size, out);
    fclose(out);
  } else {
    fprintf(stderr, "output_omake: Can't open %s.\n", filename);
  }
  
  free(p);
}

void
capture_image(char *filename)
{
	size_t len;

	len = strlen(filename);
	if (len < 5 || strncasecmp(filename + len - 4, ".png", 4) != 0)
		strncat(filename, ".png", MAX_PATH);

	if (scn_stat.image)
		write_png(scn_stat.image, filename);
}

static void
write_png(Image *image, const char *filename)
{
	png_structp pngp;
	png_infop infop;
	FILE *fp;
	int width, height;
	int i;

	fp = fopen(filename, "wb");
	if (fp == NULL) {
		fprintf(stderr, "write_png: can't create '%s'\n", filename);
		return;
	}

	width = image->ximage->width;
	height = image->ximage->height;

	pngp = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if (pngp == NULL)
		goto write_png_bad;

	infop = png_create_info_struct(pngp);
	if (infop == NULL)
		goto write_png_bad1;

	if (setjmp(png_jmpbuf(pngp)))
		goto write_png_bad1;

	png_init_io(pngp, fp);

	infop->width = width;
	infop->height = height;
	infop->compression_type = PNG_COMPRESSION_TYPE_BASE;
	infop->filter_type = PNG_FILTER_TYPE_BASE;
	infop->interlace_type = PNG_INTERLACE_NONE;
	infop->color_type = PNG_COLOR_TYPE_RGB;
	infop->bit_depth = 8;

	png_write_info(pngp, infop);

	for (i = 0; i < height; i++)
		png_write_row(pngp, image->rgbmap + i * width * 3);

	png_write_end(pngp, infop);

write_png_bad1:
	png_destroy_write_struct(&pngp, infop ? &infop : 0);
write_png_bad:
	fclose(fp);
}


typedef struct png_io_ptr_tag {
	char	*top;
	char	*curpos;
	size_t	size;
} png_io_ptr_t;

static png_io_ptr_t png_io_ptr;

static void png_read_from_memory(png_structp, png_bytep, png_size_t);

Image *
read_png(Display *d, Window w, char *pngimage, size_t size)
{
	Image *image = NULL;
	png_structp pngp;
	png_infop infop;
	int width, height, depth, color_type;
	int i;

	pngp = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if (pngp == NULL)
		goto read_png_bad;

	infop = png_create_info_struct(pngp);
	if (infop == NULL)
		goto read_png_bad1;

	if (setjmp(png_jmpbuf(pngp)))
		goto read_png_bad2;

	png_io_ptr.top = png_io_ptr.curpos = pngimage;
	png_io_ptr.size = size;
	png_set_read_fn(pngp, &png_io_ptr, png_read_from_memory);

	png_read_info(pngp, infop);

	width = infop->width;
	height = infop->height;
	depth = infop->bit_depth;
	color_type = infop->color_type;

	switch (color_type) {
	case PNG_COLOR_TYPE_RGB:
		break;
	default:
		fprintf(stderr, "Not support color_type %d\n", color_type);
		goto read_png_bad1;
	}

	image = ImageCreate(d, w, width, height);
	if (image == NULL)
		goto read_png_bad1;
	image->rgbmap = (char *)calloc(width * height * 3, 1);
	if (image->rgbmap == NULL)
		goto read_png_bad2;

	png_start_read_image(pngp);
	for (i = 0; i < height; i++)
		png_read_row(pngp, image->rgbmap + i * width * 3, 0);
	png_read_end(pngp, infop);

	png_destroy_read_struct(&pngp, &infop, 0);

	return image;

read_png_bad2:
	if (image) {
		if (image->rgbmap)
			free(image->rgbmap);
		ImageDestroy(image);
	}
read_png_bad1:
	png_destroy_read_struct(&pngp, infop ? &infop : 0, 0);
read_png_bad:
	return NULL;
}

static void
png_read_from_memory(png_structp pngp, png_bytep data, png_size_t length)
{
	png_io_ptr_t *p;
	png_size_t check;

	p = (png_io_ptr_t *)pngp->io_ptr;

	if (p->curpos + length > p->top + p->size)
		check = p->top + p->size - p->curpos;
	else
		check = length;

	memcpy(data, p->curpos, check);
	p->curpos += check;

	if (check != length)
		png_error(pngp, "Read Error");
}
